type op = Add | Sub | Mul | Div | Equal | Neq | Less | Leq | Greater | Geq | Pow | DotMul
type uop = Sin | Cos | Log | Trans | Tan

type matrix = float array array

type expr = (* Expressions *)
      Literal of float (* e.g. 42 *)
  |   MLiteral of matrix
  | 	Id of string (* e.g. foo *)
  | 	Assign of string * expr (* e.g. foo = 42 *)
  | 	Binop of expr * op * expr (* e.g. a + b *)
  | 	Call of string * expr list (* e.g. foo(1, 25) *)
  |   Uniop of uop * expr (* e.g. a^{T} *)
  |   Sum of expr * expr * expr (* e.g. /sum_{i=1}^{10}{i} *)
  |   Prod of expr * expr * expr (* e.g. /prod_{i=1}^{10}{i} *) 

type statement = (* Statements *)
	Seq of expr list (* e.g. x=3, y=4, z=5 *)

type case = {
	expression : expr;
	condition : expr;
}

type f_def = (* Formular Definition *)
	  Regular of expr (* e.g. b(m,n) = 2*n*m *)
|   Piecewise of case list (* e.g. a(m,n) = \begin{cases} \end{cases}*)

type formular = {
	fname : string;
	parameter : expr list;
	definition : f_def;
}

type program = formular list * statement list

(* AST printing utilities. Use '-a' in top-level to enable *)

(* Return a truncated string of  *)
let float2str f =
    let i = int_of_float f in
    if float_of_int i = f then string_of_int i
    else string_of_float f
;;

(* Matrix printing utilities *)
let matrix_s m =
  " [" ^ String.concat ", " (Array.to_list (Array.map (fun e -> "(" ^ (float2str e) ^ ")") m)) ^ " ]"

let traverse_matrix any = Array.to_list (Array.map (fun e -> "(" ^ matrix_s e ^ ")") any)

let matrix_decl_s m =
   "{\n         " ^ String.concat ",\n         " (traverse_matrix m) ^ "]\n}"

(* Expression/statement printing utilities *)
let rec expr_s = function
   Literal(l) -> "Literal " ^ float2str l
 | MLiteral(m) -> "MLiteral " ^ matrix_decl_s m
 | Id(s) -> "Id " ^ s
 | Binop(e1, o, e2) -> "Binop (" ^ expr_s e1 ^ ") " ^
       (match o with Add -> "Add" | Sub -> "Sub" | Mul -> "Mul" |
                     Div -> "Div" | Equal -> "Equal" | Neq -> "Neq" |
                     Less -> "Less" | Leq -> "Leq" | Greater -> "Greater" |
                     Geq -> "Geq" | Pow -> "Pow" | DotMul -> "DotMul") ^ " (" ^ expr_s e2 ^ ")"
 | Sum (down, up, e) -> "Sum (" ^ (expr_s down) ^ ")" ^ "(" ^ (expr_s up) ^ ")" ^ "(" ^ (expr_s e) ^ ")"
 | Prod (down, up, e) -> "Prod (" ^ (expr_s down) ^ ")" ^ "(" ^ (expr_s up) ^ ")" ^ "(" ^ (expr_s e) ^ ")"
 | Uniop(o, e) -> "Uniop (" ^
                    (match o with Sin -> "Sin" | Trans -> "Trans" | Cos -> "Cos" | Log -> "Log"
                      | Tan -> "Tan"
                    ) ^ " (" ^ expr_s e ^ ")"
 | Assign(v, e) -> "Assign " ^ v ^ " (" ^ expr_s e ^ ")"
 | Call(f, es) -> "Call " ^ f ^ " [" ^
        String.concat ", " (List.map (fun e -> "(" ^ expr_s e ^ ")") es) ^ "]"

let traverse any = List.map (fun e -> "(" ^ expr_s e ^ ")") any

let stmt_s = function
   Seq(ss) -> "Seq[" ^ String.concat ",\n"
                             (traverse ss) ^ "]"
(* Function priting utilities *)
let case_s c =
  "\n { expression = \"" ^ (expr_s c.expression) ^ "\"\n    condition = [" ^
  (expr_s c.condition) ^ "] }\n"

let traverse_case any = List.map (fun e -> "(" ^ case_s e ^ ")") any

let f_def_s = function
    Regular(r) -> "Regular (" ^ expr_s r ^ ")"
 | Piecewise(p) -> "Piecewise[" ^ String.concat ",\n"
                                      (traverse_case p) ^ "]"

let func_decl_s f =
  " { fname = \"" ^ f.fname ^ "\"\n   parameter = [" ^
    String.concat ", " (traverse f.parameter) ^ "]\n definition = [\n"
    ^ (f_def_s f.definition) ^ " ]}\n"
    
(* Overall program printing *)
let program_s (funcs, stats) = "([" ^ String.concat "\n" (List.map func_decl_s funcs) ^ "],\n[ " ^ String.concat "\n" (List.map stmt_s stats) ^ "])\n"
