open Ast
open Unix (* Header for get system date *)

let version = "version " ^ "0.1";;

let date =
    let months = [| "January"; "February"; "March"; "April"; "May"; "June"; "July"; "August"; "September"; "October"; "November"; "December" |]
    in
    let local_t = Unix.localtime (Unix.gettimeofday ()) in
      months.(local_t.tm_mon) ^ ", " ^ string_of_int (local_t.tm_year + 1900)
;;

(* Generate header code in c++ *)
let header v_list m_list l_count c_count fdef_latex_l title author date =
"/**
*  EZMath
*  Fast documentation and computation of math-related text
*  Created by EZMath Compiler " ^ version ^ "
**/
"
^
"
#include <iostream>
#include <fstream>
#include <vector>
#include <sstream>
#include <cmath>
#include <stdexcept>

using namespace std;
"
^
"#define NUM_OF_VARIABLES " ^ string_of_int (List.length v_list) ^ "\n" ^

"#define NUM_OF_MATRIX_VARIABLES " ^ string_of_int (List.length m_list) ^ "\n" ^

"#define NUM_OF_LOGICAL_VALIDATION " ^ string_of_int l_count ^ "\n" ^

"#define NUM_OF_FORMULAR_EVALUATION " ^ string_of_int c_count ^ "\n" ^

"#define NUM_OF_FORMULAR_DEFINITION " ^ string_of_int(List.length fdef_latex_l) ^ "\n\n" ^

"const string Title = \"" ^ "\\\\" ^ title ^ "\";\n" ^

"const string Author = \"" ^ "\\\\" ^ author ^ "\";\n" ^

"const string Date = \"" ^ date ^ "\";\n"
;;

let matrix_class =
"
class matrix
{
  //declare a vector of vectors of type double
  vector< vector<double> > s ;
  public:
    int m, n;
    //Initialize the size of s to row by col
    matrix(int row = 1, int col = 1): m(row), n(col), s(row, vector<double>(col)) {}
    matrix(int row, int col, const double data[]): m(row), n(col), s(row, vector<double>(col))
    {
      for(int i = 0; i < row; i++)
        for(int j = 0; j < col; j++)
          s[i][j] = data[i * col + j];
    }
    string printm();
  //declare the operators +,-,*,~,DotMul as friends and with return type matrix
  friend matrix operator+(const matrix&, const matrix&);
  friend matrix operator-(const matrix&, const matrix&);
  friend matrix operator*(const matrix&, const matrix&);
  friend matrix operator*(const matrix&, double);
  friend matrix operator*(double, const matrix&);
  friend matrix operator~(const matrix&);
  friend matrix DotMul(const matrix&, const matrix&);
};
string matrix::printm()
{
  stringstream lex;
  lex << \"\\\\begin {bmatrix}\" << endl;
  for(int i = 0; i < m; i++)
  {
    for(int j = 0; j < n; j++)
    {
      lex << this->s[i][j];
      lex << ((j < n - 1) ? \" & \" : ((i < m - 1)? \" \\\\\\\\\" : \"\"));
    }
    lex << endl;
  }
  lex << \"\\\\end {bmatrix}\";
  return lex.str();
}

matrix operator+(const matrix& a, const matrix& b)
{
  //declare a matrix temp to store the result and return this matrix
  if(a.m != b.m || a.n != b.n)
    throw std::runtime_error(\"Invalid matrix addition: dimensions not match\");
  matrix temp(a.m, a.n);
  for(int i = 0; i < a.m; i++)
    for(int j = 0; j < a.n; j++)
      temp.s[i][j] = a.s[i][j] + b.s[i][j];
  return temp;
}
matrix operator-(const matrix& a, const matrix& b)
{
  if(a.m != b.m || a.n != b.n)
    throw std::runtime_error(\"Invalid matrix subtraction: dimensions not match\");
  matrix temp(a.m, a.n);
  for(int i = 0; i < a.m; i++)
    for(int j = 0; j < a.n; j++)
      temp.s[i][j] = a.s[i][j] - b.s[i][j];
  return temp;
}
matrix operator*(const matrix& a, const matrix& b)
{
  if(a.n != b.m)
    throw std::runtime_error(\"Invalid matrix multiplication: dimensions not match\");
  matrix temp(a.m, b.n);
  for(int i = 0; i < a.m; i++)
  {
    for(int j = 0; j < b.n; j++)
    {
      temp.s[i][j] = 0;
      for(int k = 0; k < a.n; k++)
        temp.s[i][j] += a.s[i][k] * b.s[k][j];
    }
  }
  return temp;
}
matrix operator*(const matrix& a, double b)
{
  matrix temp(a.m, a.n);
  for(int i = 0; i < a.m; i++)
    for(int j = 0; j < a.n; j++)
      temp.s[i][j] = a.s[i][j] * b;
  return temp;
}
matrix operator*(double b, const matrix& a)
{
  matrix temp(a.m, a.n);
  for(int i = 0; i < a.m; i++)
    for(int j = 0; j < a.n; j++)
      temp.s[i][j] = a.s[i][j] * b;
  return temp;
}
matrix operator~(const matrix& trans)
{
  matrix temp(trans.n, trans.m);
  for(int i = 0; i < trans.m; i++)
    for(int j = 0; j < trans.n; j++)
      temp.s[j][i] = trans.s[i][j];
  return temp;
}
matrix DotMul(const matrix& a, const matrix& b)
{
  //declare a matrix temp to store the result and return this matrix
  if(a.m != b.m || a.n != b.n)
    throw std::runtime_error(\"Invalid matrix dot multiplication: dimensions not match\");
  matrix temp(a.m, a.n);
  for(int i = 0; i < a.m; i++)
    for(int j = 0; j < a.n; j++)
      temp.s[i][j] = a.s[i][j] * b.s[i][j];
  return temp;
}
"
;;

let dtos =
"string dtos(double d)
{
  stringstream out;
  out << d;
  return out.str();
}
"
;;

let main_preamble v_list m_list fdef_latex_l=
"
/**
* Main Function
**/
int main(int argc, char ** argv) {
"
^
"\tstring var_def[NUM_OF_VARIABLES] = {" ^ String.concat ", " (List.map (fun var -> "\"" ^ var ^ "\"") v_list) ^ "};\n" ^

"\tstring matrix_def[NUM_OF_MATRIX_VARIABLES] = {" ^ String.concat ", " (List.map (fun var -> "\"" ^ var ^ "\"") m_list) ^ "};\n\n" ^

"\tstring formular_def[NUM_OF_FORMULAR_DEFINITION] = {" ^ String.concat ", \n" (List.map (fun f -> "\"" ^  f ^ "\"" ) fdef_latex_l) ^ "};\n\n" ^

"\tstring l_result[NUM_OF_LOGICAL_VALIDATION];\n" ^

"\tstring c_result[NUM_OF_FORMULAR_EVALUATION];\n"
;;

let latex_print file =
"
  /**
    Latex Print Purpose
  */
  ofstream file(\"" ^ file ^ "\");

  //Begin
  file << \"\\\\documentclass{article}\\n\"
       << \"\\\\usepackage[utf8]{inputenc}\\n\"
       << \"\\\\usepackage{amsmath}\\n\"
       <<  Title << \"\\n\"
       <<  Author << \"\\n\"
       << \"\\\\date{\"<< Date
                   <<\"}\\n\"
       << \"\\\\begin{document}\\n\"
       << \"\\\\maketitle\\n\";

  file << \"\\\\section*{Variables Definition}\\n\";

  for (int i = 0; i < NUM_OF_VARIABLES; i++) {
    file << \"\\\\[\\n\" << var_def[i] << \" = \" << vdata[i] << \"\\n\\\\]\\n\";
  }

  file << \"\\\\section*{Matrix Definition}\\n\";
  for (int i = 0; i < NUM_OF_MATRIX_VARIABLES; i++) {
    file << \"\\\\[\\n\" << matrix_def[i] << \" = \" << mdata[i].printm() << \"\\n\\\\]\\n\";
  }

  file << \"\\\\section*{Formula Definition}\\n\";

  file << \"\\\\begin{gather*}\\n\";

  for (int i = 0; i < NUM_OF_FORMULAR_DEFINITION; i++) {
    if (i != NUM_OF_FORMULAR_DEFINITION - 1) {
      file << formular_def[i] << \"\\\\\\\\\\n\";
    }
    else {
      file << formular_def[i] << \"\\n\";
    }
  }

  file << \"\\\\end{gather*}\\n\";

  file << \"\\\\section*{Logical Validation}\\n\";

  for (int i = 0; i < NUM_OF_LOGICAL_VALIDATION; i++) {
      file << \"\\\\[\\n\" << l_result[i] << \"\\n\\\\]\\n\";
  }

  file << \"\\\\section*{Formula Evaluation}\\n\";

  for (int i = 0; i < NUM_OF_FORMULAR_EVALUATION ; i++) {
    file << \"\\\\[\\n\" <<  c_result[i] << \"\\n\\\\]\\n\";
  }

  file << \"\\\\end{document}\\n\";

  file.close();
  return 0;
}
"
;;


(* Utility Tools*)
(* *_to_string: Latex format *)
(* *_to_string_c: C++ format *)
(* *_to_string_p: Latex format for runtime-augument printing *)
(* *_to_string_interpreter: Interpreter format *)

let matrix_row_to_string_c (row: float array) : string = String.concat ", " (Array.to_list (Array.map (fun e -> float2str e) row))
;;
let matrix_to_string_c (matrix: matrix) : string = "matrix(" ^ string_of_int (Array.length matrix) ^ ", " ^ string_of_int (Array.length (matrix.(0))) ^ ", (const double []) {" ^
  (String.concat ", " (Array.to_list (Array.map matrix_row_to_string_c matrix))) ^ "})"
;;

let matrix_row_to_string (row: float array) : string = String.concat " & " (Array.to_list (Array.map (fun e -> float2str e) row))
;;
let matrix_to_string (matrix: matrix) : string =
  "\\begin{bmatrix}\n" ^
  String.concat "\\\\\n" (Array.to_list (Array.map matrix_row_to_string matrix)) ^
  "\n\\end{bmatrix}"
;;

(* expresion list to c++ string *)
let rec expr_list_to_string_c (para_list: expr list) : string = String.concat ", " (List.map (fun e -> fst(expr_to_string_c e)) para_list)

and
    expr_to_string_c (expr : expr) : string * int = match expr with
    Literal(f) -> (float2str f), 6
|   Id(s) -> s, 6
|   Assign(vid, e) -> (vid ^ "=" ^ (fst (expr_to_string_c e))), 0
|   Binop(e1, op, e2) ->
            let (s1, i1) = expr_to_string_c e1 in
            let (s2, i2) = expr_to_string_c e2 in
            (match op with
                        Add -> let (op_s, i) = "+", 3 in ((if i1 >= i then s1 else "(" ^ s1 ^ ")") ^ op_s ^ (if i2 >= i then s2 else "(" ^ s2 ^ ")")), i
                      | Sub -> let (op_s, i) = "-", 3 in ((if i1 >= i then s1 else "(" ^ s1 ^ ")") ^ op_s ^ (if i2 >= i then s2 else "(" ^ s2 ^ ")")), i
                      | Mul -> let (op_s, i) = "*", 4 in ((if i1 >= i then s1 else "(" ^ s1 ^ ")") ^ op_s ^ (if i2 >= i then s2 else "(" ^ s2 ^ ")")), i
                      | Div -> let (op_s, i) = "/", 4 in ((if i1 >= i then s1 else "(" ^ s1 ^ ")") ^ op_s ^ (if i2 >= i then s2 else "(" ^ s2 ^ ")")), i
                      | Equal -> let (op_s, i) = "==", 1 in ((if i1 >= i then s1 else "(" ^ s1 ^ ")") ^ op_s ^ (if i2 >= i then s2 else "(" ^ s2 ^ ")")), i
                      | Neq -> let (op_s, i) = "!=", 1 in ((if i1 >= i then s1 else "(" ^ s1 ^ ")") ^ op_s ^ (if i2 >= i then s2 else "(" ^ s2 ^ ")")), i
                      | Less -> let (op_s, i) = "<", 2 in ((if i1 >= i then s1 else "(" ^ s1 ^ ")") ^ op_s ^ (if i2 >= i then s2 else "(" ^ s2 ^ ")")), i
                      | Leq -> let (op_s, i) = "<=", 2 in ((if i1 >= i then s1 else "(" ^ s1 ^ ")") ^ op_s ^ (if i2 >= i then s2 else "(" ^ s2 ^ ")")), i
                      | Greater -> let (op_s, i) = ">", 2 in ((if i1 >= i then s1 else "(" ^ s1 ^ ")") ^ op_s ^ (if i2 >= i then s2 else "(" ^ s2 ^ ")")), i
                      | Geq -> let (op_s, i) = ">=", 2 in ((if i1 >= i then s1 else "(" ^ s1 ^ ")") ^ op_s ^ (if i2 >= i then s2 else "(" ^ s2 ^ ")")), i
                      | Pow -> let i = 5 in ( "pow(" ^ s1 ^ ", " ^ s2 ^ ")"), i
                      | DotMul -> let i = 4 in ( "DotMul(" ^ s1 ^ ", " ^ s2 ^ ")"), i
            )
|   Sum(a, u, e) -> let vid, ex = match a with
                        Assign(vid, ex) -> vid, ex
                      | _ -> raise (Failure ("first parameter in sum should be assign"))
                     in
                      "[](int bottom, int top){double sum=0; for(int " ^ vid ^ "=bottom; " ^ vid ^ "<top+1;" ^ vid ^ "++) sum+=" ^ (fst(expr_to_string_c e))
                      ^ "; return sum;}(" ^ "(int)(" ^ (fst(expr_to_string_c ex)) ^ "), " ^ "(int)(" ^ (fst(expr_to_string_c u)) ^ "))", 6
|   Prod(a, u, e) -> let vid, ex = match a with
                        Assign(vid, ex) -> vid, ex
                      | _ -> raise (Failure ("first parameter in prod should be assign"))
                     in
                      "[](int bottom, int top){double prod=1; for(int " ^ vid ^ "=bottom; " ^ vid ^ "<top+1;" ^ vid ^ "++) prod*=" ^ (fst(expr_to_string_c e))
                      ^ "; return prod;}(" ^ "(int)(" ^ (fst(expr_to_string_c ex)) ^ "), " ^ "(int)(" ^ (fst(expr_to_string_c u)) ^ "))", 6
|   Call(f, para_list) -> f ^ "(" ^ (expr_list_to_string_c para_list) ^ ")", 6
|   MLiteral(m) -> (matrix_to_string_c m), 6
|   Uniop(op, e) ->
            (
              match op with
                Trans -> "~" ^ "(" ^ (fst(expr_to_string_c e)) ^ ")", 6
              | Sin   -> "sin " ^ "(" ^ (fst(expr_to_string_c e)) ^ ")", 6
              | Cos   -> "cos " ^ "(" ^ (fst(expr_to_string_c e)) ^ ")", 6
              | Log   -> "log10 " ^ "(" ^ (fst(expr_to_string_c e)) ^ ")", 6
              | Tan   -> "tan" ^ "(" ^ (fst(expr_to_string_c e)) ^ ")", 6
            )
;;

let expr_list_to_string_p (seperater: string) (para_list: expr list) : string = String.concat seperater (List.map (fun e -> "dtos(" ^ fst(expr_to_string_c e) ^ ")") para_list)
;;

(* expresion list to latex string *)
let rec expr_list_to_string (para_list: expr list) : string = String.concat ", " (List.map (fun e -> fst(expr_to_string e)) para_list)

and
    expr_to_string (expr : expr) : string * int = match expr with
    Literal(f) -> (float2str f), 6
|   Id(s) -> s, 6
|   Assign(vid, e) -> (vid ^ "=" ^ (fst (expr_to_string e))), 0
|   Binop(e1, op, e2) ->
            let (s1, i1) = expr_to_string e1 in
            let (s2, i2) = expr_to_string e2 in
            (match op with
              Add -> let (op_s, i) = "+", 3 in ((if i1 >= i then s1 else "(" ^ s1 ^ ")") ^ op_s ^ (if i2 >= i then s2 else "(" ^ s2 ^ ")")), i
                      | Sub -> let (op_s, i) = "-", 3 in ((if i1 >= i then s1 else "(" ^ s1 ^ ")") ^ op_s ^ (if i2 >= i then s2 else "(" ^ s2 ^ ")")), i
                      | Mul -> let (op_s, i) = "*", 4 in ((if i1 >= i then s1 else "(" ^ s1 ^ ")") ^ op_s ^ (if i2 >= i then s2 else "(" ^ s2 ^ ")")), i
                      | Div -> let (op_s, i) = "/", 4 in ((if i1 >= i then s1 else "(" ^ s1 ^ ")") ^ op_s ^ (if i2 >= i then s2 else "(" ^ s2 ^ ")")), i
                      | Equal -> let (op_s, i) = "==", 1 in ((if i1 >= i then s1 else "(" ^ s1 ^ ")") ^ op_s ^ (if i2 >= i then s2 else "(" ^ s2 ^ ")")), i
                      | Neq -> let (op_s, i) = "!=", 1 in ((if i1 >= i then s1 else "(" ^ s1 ^ ")") ^ op_s ^ (if i2 >= i then s2 else "(" ^ s2 ^ ")")), i
                      | Less -> let (op_s, i) = "<", 2 in ((if i1 >= i then s1 else "(" ^ s1 ^ ")") ^ op_s ^ (if i2 >= i then s2 else "(" ^ s2 ^ ")")), i
                      | Leq -> let (op_s, i) = "<=", 2 in ((if i1 >= i then s1 else "(" ^ s1 ^ ")") ^ op_s ^ (if i2 >= i then s2 else "(" ^ s2 ^ ")")), i
                      | Greater -> let (op_s, i) = ">", 2 in ((if i1 >= i then s1 else "(" ^ s1 ^ ")") ^ op_s ^ (if i2 >= i then s2 else "(" ^ s2 ^ ")")), i
                      | Geq -> let (op_s, i) = ">=", 2 in ((if i1 >= i then s1 else "(" ^ s1 ^ ")") ^ op_s ^ (if i2 >= i then s2 else "(" ^ s2 ^ ")")), i
                      | Pow -> let (op_s, i) = "^", 5 in ((if i1 >= i then s1 else "(" ^ s1 ^ ")") ^ op_s ^ "{" ^ s2 ^ "}"), i
                      | _ -> raise (Failure ("No matrix in logical expressions and function calls"))
            )
|   Sum(a, u, e) -> "\\sum_{" ^ (fst(expr_to_string a)) ^ "}^{" ^ (fst(expr_to_string u)) ^ "}{" ^ (fst(expr_to_string e)) ^ "}", 1
|   Prod(a, u, e) -> "\\prod_{" ^ (fst(expr_to_string a)) ^ "}^{" ^ (fst(expr_to_string u)) ^ "}{" ^ (fst(expr_to_string e)) ^ "}", 1
|   Uniop(op, e) ->
            (
              match op with
                Sin   -> "\\sin" ^ "{" ^ (fst(expr_to_string e)) ^ "}", 1
              | Cos   -> "\\cos" ^ "{" ^ (fst(expr_to_string e)) ^ "}", 1
              | Log   -> "\\log_{10}" ^ "{" ^ (fst(expr_to_string e)) ^ "}", 1
              | Tan   -> "\\tan"      ^ "{" ^ (fst(expr_to_string e)) ^ "}", 1
              | _     -> raise (Failure ("No matrix in logical expressions and function calls"))
            )
|   Call(f, para_list) -> f ^ "(" ^ (expr_list_to_string para_list) ^ ")", 6
|   _ -> raise (Failure ("No matrix in logical expressions and function calls"))
;;

let case_to_string (case: case) : string =
  let expression = fst(expr_to_string case.expression) in
  let condition = fst(expr_to_string case.condition) in
  expression ^ " & " ^ condition
;;

let case_list_to_string (case_list: case list) : string = String.concat "\\\\\n" (List.map case_to_string case_list)
;;

let formular_to_string (formular : formular) : string =
  let fname = formular.fname in
  let parameter = expr_list_to_string formular.parameter in
  let definition = match formular.definition with
             Regular(e) -> fst(expr_to_string e)
          |  Piecewise(case_list) -> "\\begin{cases}\n" ^ (case_list_to_string case_list) ^ "\n\\end{cases}"
  in
  fname ^ "(" ^ parameter ^ ")=" ^ definition
;;

let case_to_string_interpreter (case: case) : string =
  let expression = fst(expr_to_string case.expression) in
  let condition = fst(expr_to_string case.condition) in
  "    " ^ expression ^ ", if " ^ condition ^ "."
;;

let case_list_to_string_interpreter (case_list: case list) : string = String.concat " Or\n" (List.map case_to_string_interpreter case_list)
;;

let formular_to_string_interpreter (formular : formular) : string =
  let fname = formular.fname in
  let parameter = expr_list_to_string formular.parameter in
  let definition = match formular.definition with
             Regular(e) -> fst(expr_to_string e)
          |  Piecewise(case_list) -> "{\n" ^ (case_list_to_string_interpreter case_list) ^ "\n}"
  in
  fname ^ "(" ^ parameter ^ ") = " ^ definition
;;
