%{
    open Ast
    (* Convert matrix in 2d list into 2d array *)
    let list2matrix data =
        let m = List.length data in
        if m <= 0 then raise (Failure ("Parse with matrix row == 0")) else
        let n = List.length (List.hd data) in
        if n <= 0 then raise (Failure ("Parse with matrix col == 0")) else
        Array.of_list(
            List.rev ((List.fold_left (fun l row -> (
                if ((List.length row) <> n) then raise (Failure ("row item not consistent"))
                else Array.of_list row)::l) [] data))
        )
%}


/* Symbols */
%token SEMI LPAREN RPAREN LBRACE RBRACE COMMA BEGIN END CASES BMATRIX CARET DBACKSLASH AND EOF
/* Float Operation */
%token ASSIGN EQUAL NEQ GEQ LEQ GT LT PLUS MINUS TIMES DIVIDE
%token SUM PROD UNDERSCORE
%token SIN COS LOG TAN
/* Matrix Binop */
%token DOTMUL TRANS
%token <float> FLOAT
%token <string> VID /* Variable ID both for float variable and matrix */
%token <string> FID /*Formular ID */

/*Precedence and Association */

%right ASSIGN
%left EQUAL NEQ
%left LT GT LEQ GEQ
%left PLUS MINUS
%left TIMES DIVIDE DOTMUL
%left CARET TRANS


%start program
%type <Ast.program> program

%%

program:
        /* nothing */     { [], [] }
    |   program fdecl     { ($2 :: fst $1), snd $1  } /* Prepend function decl */
    |   program stmt      {  fst $1, ($2 :: snd $1) } /* Prepend statement */

/* For Matrix Data */
row_list :
        row  { [List.rev $1] }
    |   row_list DBACKSLASH row { (List.rev $3) :: $1 }

row :
        FLOAT   { [$1] }
    |   MINUS FLOAT { [-. $2] }
    |   row AND FLOAT { $3 :: $1 }
    |   row AND MINUS FLOAT { (-. $4)::$1 }

fdecl:
    FID arguments RPAREN ASSIGN forumular_def
    {
        {
            fname = $1;
            parameter = $2;
            definition = $5;
        }
    }
/* Regular function: single expr. Piecewise: match cases with exprs */
forumular_def :
        regular_def     { $1 }
    |   piecewise_def   { $1 }

regular_def :
    expr                { Regular($1) }

piecewise_def :
    BEGIN LBRACE CASES RBRACE case_list END LBRACE CASES RBRACE { Piecewise(List.rev $5) }

case_list :
        case                      { [$1] }
    |   case_list DBACKSLASH case { $3 :: $1 }

case :
    expr AND expr
    {
        {
            expression = $1;
            condition = $3;
        }
    }

/* Top-level expression */
expr :
    /* float constant*/
        FLOAT                           { Literal($1) }
    /* matrix constant */
    |   BEGIN LBRACE BMATRIX RBRACE row_list END LBRACE BMATRIX RBRACE
                                        { MLiteral(list2matrix(List.rev $5)) }
    |   VID                             { Id($1) }
    |   exec_expr                       { $1 }

/* Expression allowed tp appear alone in statement */
exec_expr:
        expr PLUS expr                  { Binop($1, Add, $3) }
    |   expr MINUS expr                 { Binop($1, Sub, $3) }
    |   expr TIMES expr                 { Binop($1, Mul, $3) }
    |   expr DIVIDE expr                { Binop($1, Div, $3) }
    |   expr EQUAL expr                 { Binop($1, Equal, $3) }
    |   expr NEQ expr                   { Binop($1, Neq, $3) }
    |   expr LT expr                    { Binop($1, Less, $3) }
    |   expr LEQ expr                   { Binop($1, Leq, $3) }
    |   expr GT expr                    { Binop($1, Greater, $3) }
    |   expr GEQ expr                   { Binop($1, Geq, $3) }
    |   expr CARET LBRACE expr RBRACE   { Binop($1, Pow, $4) }
    |   SUM UNDERSCORE LBRACE expr RBRACE CARET LBRACE expr RBRACE LBRACE expr RBRACE { Sum($4, $8, $11) }
    |   PROD UNDERSCORE LBRACE expr RBRACE CARET LBRACE expr RBRACE LBRACE expr RBRACE { Prod($4, $8, $11) }
    |   VID ASSIGN expr                 { Assign($1, $3) }
    |   FID arguments RPAREN           { Call($1, $2) }
    |   LPAREN expr RPAREN              { $2 }
    |   LPAREN MINUS expr RPAREN        { Binop(Literal(0.), Sub, $3)}
    |   expr DOTMUL expr                { Binop($1, DotMul, $3) }
    |   expr CARET LBRACE TRANS RBRACE  { Uniop(Trans, $1)  }
    |   SIN LBRACE expr RBRACE          { Uniop(Sin, $3)    }
    |   COS LBRACE expr RBRACE          { Uniop(Cos, $3)    }
    |   LOG LBRACE expr RBRACE          { Uniop(Log, $3)    }
    |   TAN LBRACE expr RBRACE          { Uniop(Tan, $3)    }

arguments:
                                       { [] }
    |   argument_list                    { List.rev $1 }

argument_list :
        expr                           { [$1] }
    |   argument_list COMMA expr         { $3 :: $1 }

/* Only exec_expr allowed alone in statement */
stmt:
    exec_expr_list                     { Seq(List.rev $1) }

exec_expr_list :
        exec_expr                      { [$1] }
    |   exec_expr_list COMMA exec_expr { $3 :: $1 }
